/*
 * include/linux/dipc.h
 *
 * Copyright (C) Kamran Karimi
 */
#ifndef _LINUX_DIPC_H
#define _LINUX_DIPC_H 
#include <linux/sched.h>

extern int dipc(int first, int cmd, void *ptr);

#define DIPC_VERSION "0.15"
 
#define SHM 0
#define MSG 1
#define SEM 2

#define DIPC_SUCCESS 0
#define DIPC_FAIL -1

#define DIPC_REMPROT  2 /* remove protection bits */
#define DIPC_DEFPROT  3 /* set protection to default */

#define DIPC_SLEEP    6 /* sleep till there is work */

#define DIPC_SETMNG  10 /* set the dsm manager task */
#define DIPC_REMMNG  11 /* remove the dsm manager */
#define DIPC_GETWORK 13 /* get a work */
#define DIPC_RETWORK 14 /* return the result of the work done */

#define DIPC_SHM_INFO 20 /* get info about a structure */
#define DIPC_SEM_INFO 21
#define DIPC_MSG_INFO 22

#define DISABLE_RD 50
#define DISABLE_WR 51

#define WANTRD     61
#define WANTWR     62
#define DETACH     63
#define AK         64

#define WRPROT_SND 72
#define RDPROT     73
#define RDPROT_SND 74 /* first read-protect, then send the contents */
#define NOP_SND    75 /* nop, just send the contents to third party */
#define RCV        76
#define TSTGET     77

#define RMID       81
#define DELKEY     82
#define SEARCH     83
#define COMMIT     84

#define NOT_FOUND  0
#define FOUND_IPC  1
#define FOUND_DIPC 2

struct dipc_info
{
 int key;
 int type; /* SHM, SEM or MSG */
 int pid; /* pid of the requesting task */
 int uid; /* uid of the requesting task */
 int func; /* what to do? (MSGCTL, SEMOP,...) */
 int arg1; /* arg1 to arg4 are the original arguments */
 int arg2; /* some of them may not be used */
 int arg3;
 int arg4;
 int dsize; /* is there any data? */
 int owner; /* network address of the owner */
};

struct dipc_request
{
 struct dipc_info info;
 int working; /* shows if this request is already in works */
 char *data; /* data for this request */
 struct wait_queue *lock; /* requesting task will sleep here */
 struct dipc_request *next, *prev;
};

extern int dipc_present(void);
extern int dipc_manager(struct task_struct *task);

extern int dipc_search_key(key_t key,int type,int distrib,int *size);
extern int dipc_commit_key(key_t key, int type, int size, int mode);
extern int dipc_test_key(key_t key, int type, int size, int flag);
extern int dipc_del_key(key_t key, int type);
extern int dipc_rmid_key(key_t key, int type);

extern void dipc_add_request(struct dipc_request *dr);

extern void dipc_excl(key_t key, int type);
extern int dipc_check(key_t key, int type);
extern void dipc_wake(key_t key, int type);


/* the rest should be accessible to user programs */
#define DIPC_SIG_WRITER  SIGPWR
#define DIPC_SIG_READER  SIGURG

/* repeat the number in all bytes, so that endian-ness is avoided */
#define LINUX_i386 0x01010101
#define LINUX_68k  0x02020202

#define MY_DIPC_ARCH LINUX_i386

#endif /* _LINUX_DIPC_H */
