/*
 * dipcker.c
 *
 * Allows the user to see and delete some information about IPC structures
 * in the local kernel.
 *
 * By Kamran Karimi
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <errno.h>
#include <linux/ipc.h>
#include <linux/shm.h>
#include <linux/sem.h>
#include <linux/msg.h>
#include <sys/shm.h>
#include <sys/sem.h>
#include <sys/msg.h>
#include <arpa/inet.h>
#include <linux/dipc.h>

struct info_link
{
 int type;
 union
 {
  struct shmid_ds shm;
  struct semid_ds sem;
  struct msqid_ds msg;
 } info;
 struct info_link *next, *prev;
}; 

struct info_link first_link;
struct info_link *last_link;

void main(int argc, char **argv)
{
 int key, mode, shmid, semid, msgid;
 int foundshm, foundsem, foundmsg;
 struct shmid_ds shm;
 struct semid_ds sem;
 struct msqid_ds msg;
 struct in_addr addr;
 char *errch;
 struct info_link *il;

 if(argc < 2)
 {
  printf("USAGE:\n       %s info\n", argv[0]);
  printf("       %s rm <type> <key>\n", argv[0]);
  exit(0);
 }
 first_link.next = first_link.prev = NULL;
 first_link.type = -1;
 last_link = &first_link;
 if(!strcasecmp(argv[1], "info"))
 {  
  for(key = 0; key < SHMMNI; key++)
  {
   shmid = dipc(key, DIPC_SHM_INFO, &shm);
   if(shmid != -1)
   {
    il = malloc(sizeof(struct info_link));
    if(il == NULL)
    {
     fprintf(stderr,"dipcker: Out of memory\n");
     break;
    }
    il->type = SHM;
    il->info.shm = shm;
    il->next = NULL;
    il->prev = last_link;
    last_link->next = il;
    last_link = il;
   }
  }
  for(key = 0; key < MSGMNI; key++)
  {
   msgid = dipc(key, DIPC_MSG_INFO, &msg);
   if(msgid != -1) 
   {
    il = malloc(sizeof(struct info_link));
    if(il == NULL)
    {
     fprintf(stderr,"dipcker: Out of memory\n");
     break;
    }
    il->type = MSG;
    il->info.msg = msg;
    il->next = NULL;
    il->prev = last_link;
    last_link->next = il;
    last_link = il;
   }
  }
  for(key = 0; key < SEMMNI; key++)
  {
   semid = dipc(key, DIPC_SEM_INFO, &sem);
   if(semid != -1)
   {
    il = malloc(sizeof(struct info_link));
    if(il == NULL)
    {
     fprintf(stderr, "dipcker: Out of memory\n");
     break;
    }
    il->type = SEM;
    il->info.sem = sem;
    il->next = NULL;
    il->prev = last_link;
    last_link->next = il;
    last_link = il;
   }
  }
  foundshm = foundsem = foundmsg = 0;
  il = first_link.next;
  while(il)
  {
   if(il->type != il->prev->type)
   {
    if(il->type == SHM)
    {
     foundshm = 1;
     printf("\nShared Memories:\n");
     printf("    Key   DIPC  Attachs     Size   Read  Write   Owner\n");
    }
    else if(il->type == MSG)
    {
     foundmsg = 1;
     printf("\nMessage Queues:\n");
     printf("    Key   DIPC  Messages    Size  Owner\n");
    }    
    else if(il->type == SEM)
    {
     foundsem = 1;
     printf("\nSemaphore Sets:\n");
     printf("    Key   DIPC  Semaphs   Owner\n");
    }    
   }
   if(il->type == SHM)
   {
    shm = il->info.shm;
    addr.s_addr = shm.shm_perm.seq; 
    printf(" %6d   %4s   %6d   %6d   %4s   %4s   %-15.15s\n",
          shm.shm_perm.key, 
          (shm.shm_perm.mode & IPC_DIPC) ? "Yes" : "No",
          shm.shm_nattch, shm.shm_segsz,  
          (shm.shm_perm.mode & IPC_ABS) ? "NO" : "YES", 
          (shm.shm_perm.mode & IPC_WRP) ? "NO" : "YES",
          (addr.s_addr == -1) ? "N/A" : inet_ntoa(addr));
   }
   else if(il->type == MSG)
   {
    msg = il->info.msg;
    addr.s_addr = msg.msg_perm.seq;
    printf(" %6d   %4s   %6d   %6d  %-15.15s\n",
            msg.msg_perm.key, 
            (msg.msg_perm.mode & IPC_DIPC) ? "Yes" : "No",
            msg.msg_qnum, msg.msg_cbytes, 
            (addr.s_addr == -1L) ? "N/A" : inet_ntoa(addr));
   }   
   else if(il->type == SEM)
   {
    sem = il->info.sem; 
    addr.s_addr = sem.sem_perm.seq;
    printf(" %6d   %4s   %6d   %-15.15s\n", 
            sem.sem_perm.key, 
            (sem.sem_perm.mode & IPC_DIPC) ? "Yes" : "No",
            sem.sem_nsems, 
            (addr.s_addr == -1L) ? "N/A" : inet_ntoa(addr));
   }
   il = il->next;   
  }
  if(foundshm == 0)
   printf("\nNo shared memory was found in this computer\n");
  if(foundmsg == 0)
   printf("\nNo message queue was found in this computer\n");
  if(foundsem == 0)
   printf("\nNo semaphore set was found in this computer\n");
  
  il = first_link.next;
  while(il)
  {
   last_link = il->next;
   free(il);
   il = last_link;
  }
 }
 
 /******  removal ******/
 else if(!strcasecmp(argv[1], "rm"))
 { 
  key = strtol(argv[3], &errch, 10);
  if(errch == argv[3])
  {
   fprintf(stderr,"dipcker: ERROR. Invalid Key\n");
   exit(20);
  }
  if(!strcasecmp(argv[2], "shm"))
  {
   shmid = dipc(key, DIPC_SHM_INFO, &shm);
   if(shmid != -1)
   {
    mode = (shm.shm_perm.mode & IPC_DIPC) ? IPC_DIPC : 0;
    shmid = shmget(key, 0, mode);
    if(shmid < 0)
    {
     fprintf(stderr, 
       "dipcker: shmget() failed, BECAUSE: %s\n", strerror(errno));
     exit(20);
    }
    if(shmctl(shmid, IPC_RMID, NULL) < 0)
    {
     fprintf(stderr,
       "dipcker: shmctl() failed, BECAUSE: %s\n", strerror(errno)); 
     exit(20);
    }
    printf("shared memory was successfully removed\n");
   }
   else
   {
    fprintf(stderr, "dipcker: Can not find the shared memory\n");
    exit(20);
   }
  }
  else if(!strcasecmp(argv[2], "msg"))
  {
   msgid = dipc(key, DIPC_MSG_INFO, &msg);
   if(msgid != -1)
   {
    mode = (msg.msg_perm.mode & IPC_DIPC) ? IPC_DIPC : 0;
    msgid = msgget(key, mode);
    if(msgid < 0)
    {
     fprintf(stderr,
      "dipcker: msgget() failed, BECAUSE: %s\n", strerror(errno));
     exit(20);
    } 
    if(msgctl(msgid, IPC_RMID, NULL) < 0)
    {
     fprintf(stderr,"dipcker: msgctl() failed, BECAUSE: %s\n", strerror(errno));
     exit(20);
    }
    printf("message queue was successfully removed\n");
   }
   else
   {
    fprintf(stderr, "dipcker: Can not find the message queue\n");
    exit(20);
   }
  }
  else if(!strcasecmp(argv[2], "sem"))
  {
   semid = dipc(key, DIPC_SEM_INFO, &sem);
   if(semid != -1)
   {
    mode = (sem.sem_perm.mode & IPC_DIPC) ? IPC_DIPC : 0;
    semid = semget(key, 0, mode);
    if(semid < 0)
    {
     fprintf(stderr, 
       "dipcker: semget() failed, BECAUSE: %s\n", strerror(errno));
     exit(20);
    }
    if(semctl(semid, 0, IPC_RMID, NULL) < 0)
    {
     fprintf(stderr,"dipcker: semctl() failed, BECAUSE: %s\n", strerror(errno));
     exit(20);
    }
    printf("semaphore set was successfully removed\n");
   }
   else
   {
    fprintf(stderr, "dipcker: Can not find the semaphore set\n");
    exit(20);
   }
  } 
  else
   fprintf(stderr, "dipcker: Unknown IPC structure\n");  
 }
 else
  fprintf(stderr,"dipcker: Unknown command\n");
}
